/*
** Module   : ClientDriver.java
** Abstract : CHUI client command line application driver
**
** Copyright (c) 2005-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 NVS 20050827   @22390 Created this production level client driver.
** 002 NVS 20060110   @23849 Now the omission of the application key is
**                           allowed and causes StandardClient to be used.
** 003 GES 20060815   @28609 Refactored to remove duplicate code and share
**                           a common base class.
** 004 NVS 20061017   @30459 Switched to using connect() method instead of
**                           a callback approach.
** 005 GES 20061215   @31705 Changes in bootstrap config interface.
** 006 GES 20070111   @31780 Hard coded CHUI client processing here and
**                           removed some unnecessary processing/dead
**                           code.
** 007 GES 20070112   @31811 Removed initial debug level (it wasn't being
**                           used downstream) and added application args.
** 008 GES 20070115   @31826 Added support for conversation mode.
** 009 EVL 20070609   @34005 Adding explicit import of the class
**                           com.goldencode.p2j.net.Queue to eliminate
**                           conflict with the same class from java.util
**                           package to be able to compile for Java 6.
** 010 ECF 20071101   @35890 Refactored to accommodate net package mods.
**                           Eliminated ClientBootstrap dependencies,
**                           integrated SessionManager.
** 011 GES 20080409   @37911 Initialize remote directory access after the
**                           session is started.
** 012 GES 20090720   @43239 Better error handling for throwables.
** 013 GES 20090723   @43356 More error handling improvements and docs.
** 014 GES 20091103   @44304 Added a logging facility for detailed debugging.
** 015 GES 20110928          Moved core logic into a common class so it can
**                           be easily shared.
** 016 MAG 20140131          For web clients redirect the stderr to a file.
*/

package com.goldencode.p2j.main;

import java.io.*;

import com.goldencode.p2j.cfg.*;

/**
 * General purpose command line driver to start a client process based on a
 * bootstrap configuration file and/or command line parameters. 
 * <p>
 * This driver performs a single logon to the server, invokes the application
 * class and exits, when the latter is done.
 * <p>
 * Configuration file can be plain (unencrypted) XML file or encrypted XML. 
 * If no password is given with the command line, the file is assumed to be
 * plain XML.
 * <p>
 * Zero, one or two configuration files can be given as parameters along 
 * with their respective passwords. If not specified, it will default to
 * "standard_client.xml" in the current directory.
 * <p>
 * The two file approach is recommended as it allows for a natural way of
 * separating common or related configuration information (for all users or
 * a group of users, normally supplied in the first file), from the 
 * user-specific configuration information (usually supplied in the second
 * file). By merely specifying a different second file, another subject can
 * use the application while sharing the first configuration file. 
 * <p>
 * All entries in the bootstrap configuration can be overridden by a command
 * line parameter using the "category:group:key=value" syntax (see below).
 * <p>
 * The bootstrap configuration can specify a custom class as the server
 * process by using the "applications:class:server" key.  If not specified
 * in the configuration file or on the command line, its value will be 
 * defaulted to {@link StandardServer}.
 * <p>
 * The full syntax of the command line is:
 * <p>
 * <code>
 * java ClientDriver [-?]
 *                   [<i>config1</i>]
 *                   [-p <i>config1_file_password</i>]
 *                   [<i>config2</i>]
 *                   [-p <i>config2_file_password</i>]
 *                   [-d <i>initial_debuglevel</i>]
 *                   [[[category:]group:]key=value]...  
 *                   [-a application_arguments]
 * </code>
 * <p>
 * where arguments can be specified in any order and the application, except
 * for the "application_arguments" which can only appear at the end of the
 * command line.  The first password specified will be used with the first
 * config file specified and likewise with the second password and config
 * file respectively.
 * <p>
 * '-?' requests syntax help and exits.
 * <p>
 * Password can be given with the command line or entered interactively. 
 * One special value can be used: "?" means interactive password input.
 * <p>
 * Arguments in the form "category:group:key=value" are bootstrap 
 * configuration overrides. They may be used to add or change values of the
 * configuration keys in the specified category and group. A short form of
 * override (one that omits the category and group) is allowed after a long
 * form override and refers to the most recently specified category and 
 * group.
 */
public class ClientDriver
extends CommonDriver
{
   {
      // initialize our protected members
      
      defCfg = "standard_client.xml";
      maxCfg = 2;
      
      helptxt = new String[]
      {
         "-----------------------P2J Client Driver-----------------------",
         "",
         "Syntax:",
         "",
         "java ClientDriver -?",
         "                  [config1]",
         "                  [-p config1_file_password]",
         "                  [config2]",
         "                  [-p config2_file_password]",
         "                  [-d initial_debug_level]",
         "                  [parameter_overrides]",
         "                  [-a application_arguments]",
         "",
         "Where:",
         "",
         "-?                    = displays this help screen",
         "config                = a relative or absolute filename of a valid",
         "                        bootstrap configuration file for this",
         "                        client",
         "config_file_password  = the password to use to decrypt the config", 
         "                        file or '?' to force an interactive",
         "                        password prompt",
         "initial_debug_level   = an integer value that specifies the",
         "                        logging that should occur before server's",
         "                        main configuration (in the directory) can",
         "                        be read",
         "parameter_overrides   = configuration values that add to or",
         "                        replace values that can be specified in",
         "                        the bootstrap configuration; the override",
         "                        must take one of three forms:",
         "                           category:group:key=value",
         "                           group:key=value",
         "                           key=value",
         "                        (the shorter forms will use the most",
         "                        recently specified category and/or group",
         "                        as needed)",
         "application_arguments = a list of application-specific strings",
         "                        which will be passed to the server class'",
         "                        arguments() method",
         "",
         "All arguments can be specified in any order except '-?' which must", 
         "always appear first and '-a' which must always appear last. The",
         "first password specified will be used with the first config",
         "file specified and likewise with the second password and config",
         "file respectively."
      };
   }

   /**
    * Allows a subclass to check if the given configuration matches any
    * required values.  Edits may be made if needed.
    *
    * @param    bc
    *           The configuration to be checked.
    */
   protected void checkConfig(BootstrapConfig bc)
   {
      bc.setServer(false);
      
      String convo = null;
      
      try
      {
         convo = bc.getConfigItem("net", "queue", "conversation");
      }
      
      catch (ConfigurationException ce)
      {
         // ignore
      }
         
      // default to conversation mode if not specified
      if (convo == null)
      {
         bc.setConfigItem("net", "queue", "conversation", "true");
      }
      
      // either way, force start thread to false (this client is going to
      // use its main thread as the conversation thread so we don't need
      // to start one
      bc.setConfigItem("net", "queue", "start_thread", "false");
   }
   
   /**
    * Runs the primary application using the preprocessed configuration and
    * target object.
    * 
    * @param    config
    *           The configuration to use.
    * @param    args
    *           Application-specific command line arguments.  All
    *           standardized driver arguments or configuration related
    *           arguments are removed.
    */
   protected void start(BootstrapConfig config, String[] args)
   throws Exception
   {
      ClientCore.start(config, true, args);
      
      // ensure that remaining non-daemon threads don't stop us from closing
      System.exit(0);
   }
   
   /**
    * For web clients redirect the log messages to a file.
    * 
    * @param   args
    *          Process command line arguments.
    */
   public static void setWebClientLog(String[] args)
   {
      for (String arg : args)
      {
         if ("client:driver:type=chui_web".equals(arg))
         {
            // build log file name
            String logFile = String.format("client_%s_%d.log", 
                                           System.getProperty("user.name"), 
                                           System.currentTimeMillis());
            try
            {
               // redirect stderr
               System.setErr(new PrintStream(logFile));
            }
            catch (FileNotFoundException e)
            {
               // normally should not occur.
            }

            break;
         }
      }      
   }
   
   /**
    * The command line entry point.
    * 
    * @param     args
    *            The array of command-line parameters.
    *
    * @throws    ConfigurationException
    *            In case of problems with the configuration.
    */
   public static void main(String[] args)
   {
      // for web clients only redirect stderr to a log file
      setWebClientLog(args);
      // start processing
      ClientDriver client = new ClientDriver();
      client.process(args);
   }
}
